#include <parallel.h>

void transmit_intro_letter(ZIP_MAILER *Mailer,
                           long Row_Dis_Type, long Col_Dis_Type,
                           long Row_Extra_param1, long Row_Extra_param2,
                           long Col_Extra_param1, long Col_Extra_param2,
                           long Strategy)
{
#define NLONGS 11

   char *Letter;
   long *Header;

   int Letter_Size = NLONGS * sizeof(long);

#ifdef DEBUG_TRANSMIT_INTRO
   print("transmit_intro_letter(): entry...");
#endif

   Zip_push(Mailer);

   Letter    = Zip_malloc(Letter_Size);
   Header    = (long *) Letter;
   Header[0] = INTRO_LETTER;
   Header[1] = (long) Letter_Size;
   Header[2] = NLONGS;
   Header[3] = 0L;

   Header[4] = (long)Row_Dis_Type;
   Header[5] = (long)Col_Dis_Type;

   Header[6] = Row_Extra_param1;
   Header[7] = Row_Extra_param2;
   Header[8] = Col_Extra_param1;
   Header[9] = Col_Extra_param2;

   Header[10] = Strategy;

   ctohl(Header, NLONGS);

   G2_send(Letter, 0, 0);

   Zip_pop();
#undef NLONGS
}

void receive_intro_letter(ZIP_MAILER *Mailer,
                          short *Row_Dis_Type, short *Col_Dis_Type,
                          long *Row_Extra_param1, long *Row_Extra_param2,
                          long *Col_Extra_param1, long *Col_Extra_param2,
                          long *Strategy)
{
   int   p;
   int   q;

   char *Letter;
   int   Letter_Size;

   long *Header;
   long *Dis_Header;
   long *Mat_Data_Header;

   int   G2_pm;
   int   serror;

   Zip_push(Mailer);

   G2_pq(p, q);

   G2_pm = (!p) && (!q);

   if (G2_pm)
   {
#ifdef DEBUG_RECEIVE_INTRO
      print("I am G2 PM; waiting for letter from HOST");
#endif
      Letter      = (char *) G2_recvb(-1, -1);
      Header      = (long *)Letter;
      Letter_Size = Header[1];
   }
   else
   {
      Letter_Size = 0;
      Letter      = NULL;
   }

   /* fanout: */
   serror = G2_fanout(&Letter, &Letter_Size, 0, 0);
   Header = (long *) Letter;

#ifdef DEBUG_RECEIVE_INTRO
   print("receive_intro_letter(): After fanout\n");
#endif

   if (*Header != INTRO_LETTER)
   {
      print("csparsetest1a: anticipated INTRO_LETTER=%ld, received %ld",
            (long)INTRO_LETTER, *Header);

      exit(1);
   }

   Dis_Header   = Header + 4;
   *Row_Dis_Type = (short) Dis_Header[0];
   *Col_Dis_Type = (short) Dis_Header[1];

   Mat_Data_Header = (Dis_Header + 2);

   /* data for one and two-parameter distributions */
   *Row_Extra_param1 = Mat_Data_Header[0]; /* "B" */
   *Row_Extra_param2 = Mat_Data_Header[1]; /* "S" */
   *Col_Extra_param1 = Mat_Data_Header[2]; /* "B" */
   *Col_Extra_param2 = Mat_Data_Header[3]; /* "S" */

   *Strategy = Mat_Data_Header[4];

   if (G2_pm)
      Zip_free(Letter);
   else
      free(Letter);

   Zip_pop();
}

void transmit_matrix_letter(ZIP_MAILER *Mailer, matrix_host_bin *Matrix)
{
#define NLONGS 8

   char      *Subroutine_name = "transmit_matrix_letter():";
   char     *Letter;
   int       Letter_Size;

   long     *Header;

   long     *Matrix_i;
   long     *Matrix_j;

   double   *Matrix_Value;

   int       M = Matrix -> M;
   int       N = Matrix -> N;

   int       P = Matrix -> P;
   int       Q = Matrix -> Q;

   int       p;
   int       q;

   int     **A_element_Count = Matrix -> A_element_Count;
   long      Matrix_element_Count;
   int       Index;
   element ***A = Matrix -> A;
   element   *Pointer;

#ifdef DEBUG_TRANSMIT_MATRIX
   print("transmit_matrix_letter(): entry...");
#endif

   Zip_push(Mailer);

   for (p = 0; p < P; p++)
      for (q = 0; q < Q; q++)
      {
         Matrix_element_Count = A_element_Count[p][q];

         Letter_Size = NLONGS * sizeof(long)
                       + Matrix_element_Count
                         * ((2 * sizeof(long)) + sizeof(double));

         Letter    = Zip_malloc(Letter_Size);
         Header    = (long *) Letter;
         Header[0] = MATRIX_LETTER;
         Header[1] = (long)Letter_Size;
         Header[2] = NLONGS;
         Header[3] = 0L;

         Header[4] = M;
         Header[5] = N;

         Header[6] = Matrix_element_Count;
         Header[7] = 0L;

         htocl(Header, NLONGS);

#ifdef DEBUG_transmit_matrix_letter
         print("transmit_matrix_letter(): Matrix_element_Count = %d\n",
               Matrix_element_Count);
#endif
         if (Matrix_element_Count > 0)
         {
            Matrix_i = &(Header[NLONGS]);
            Matrix_j = &(Matrix_i[Matrix_element_Count]);
            Matrix_Value = (double *) &(Matrix_j[Matrix_element_Count]);

            Index = 0;
            Pointer = A[p][q];
            do
            {
               Matrix_i[Index] = Pointer -> i;
               Matrix_j[Index] = Pointer -> j;
               Matrix_Value[Index] = Pointer -> Value;
               Index++;
               Pointer = Pointer -> Next;
            } while (Pointer != NULL);

            if (Index < Matrix_element_Count)
            {
               print("%s Lost matrix entries\n", Subroutine_name);
               print("Index = %d, Matrix_element_Count = %d\n", Index,
                     Matrix_element_Count);
            }
            htocl(Matrix_i, Matrix_element_Count);
            htocl(Matrix_j, Matrix_element_Count);
            htocd(Matrix_Value, Matrix_element_Count);
         }

         G2_send(Letter, p, q);
      }

   Zip_pop();

#undef NLONGS
}

void receive_matrix_letter(ZIP_MAILER *Mailer, matrix_cube_bin *Matrix)
{
   char   *Subroutine_name = "receive_matrix_letter():";
   int     p;
   int     q;

   char   *Letter;

   long   *Header;
   long   *Mat_Data_Header;
   long   *Matrix_i;
   long   *Matrix_j;
   double *Matrix_Value;

   long    Matrix_element_Count;

   Zip_push(Mailer);

   G2_pq(p, q);

   Letter      = (char *) G2_recvb(-1, -1);

   Header      = (long *)Letter;

   if (*Header != MATRIX_LETTER)
   {
      print("csparsetest1a: anticipated MATRIX_LETTER=%ld, received %ld",
            (long)MATRIX_LETTER, *Header);

      exit(1);
   }

   Mat_Data_Header = Header + 4;
   Matrix_i = Mat_Data_Header + 4;
   Matrix_j = Matrix_i + Mat_Data_Header[2];
   Matrix_Value = (double *) (Matrix_j + Mat_Data_Header[2]);

   Matrix -> M = Mat_Data_Header[0];
   Matrix -> N = Mat_Data_Header[1];

   if (Matrix -> Array_len < Mat_Data_Header[2])
   {
      if (Matrix -> Array_len > 0)
      {
         free(Matrix -> i);
         free(Matrix -> j);
         free(Matrix -> Value);
      }

      Matrix -> Array_len = Mat_Data_Header[2];
      Matrix -> i = (long *) malloc(Mat_Data_Header[2] * sizeof(long));
      Matrix -> j = (long *) malloc(Mat_Data_Header[2] * sizeof(long));
      Matrix -> Value = (double *) malloc(Mat_Data_Header[2] * sizeof(double));

      if (Matrix -> Value == NULL)
      {
         print("%s Out of memory!\n", Subroutine_name);
         exit(1);
      }
   }

   Matrix -> num_element = Mat_Data_Header[2];

   bcopy(Matrix_i, Matrix -> i, Mat_Data_Header[2] * sizeof(long));
   bcopy(Matrix_j, Matrix -> j, Mat_Data_Header[2] * sizeof(long));
   bcopy(Matrix_Value, Matrix -> Value, Mat_Data_Header[2] * sizeof(double));

   Zip_free(Letter);

   Zip_pop();
}

void transmit_timing_letter(ZIP_MAILER *Mailer, tbx_timer *Timer)
{
#define NLONGS 6

   char   *Subroutine_name = "transmit_timing_letter():";

   int     p;
   int     q;
   
   char   *Letter;
   int     Letter_Size;

   long    Timing[2];

   long   *Header;
   long   *Payload;

   Method  timing_max_method = {(int (*)())timing_max_fn, NULL};

#ifdef DEBUG_transmit_timing_letter
   print("transmit_timing_letter(): entry...");
#endif

   Zip_push(Mailer);

   G2_pq(p, q);

   Timing[0] = Timer -> total_sec;
   Timing[1] = Timer -> total_usec;

#ifdef DEBUG_transmit_timing_letter
   print("before combine Timing[0] = %ld, Timing[1] = %ld\n", Timing[0],
          Timing[1]);
#endif

   G2_combine(Timing, timing_max_method, 2*sizeof(int), 1);

#ifdef DEBUG_transmit_timing_letter
   print("after combine Timing[0] = %ld, Timing[1] = %ld\n", Timing[0],
          Timing[1]);
#endif

   if ((p == 0) && (q == 0))
   {
      Letter_Size = NLONGS * sizeof(long);
      Letter      = Zip_malloc(Letter_Size);
      Header      = (long *) Letter;
      Payload     = Header + 4;

      Header[0] = TIMING_LETTER;
      Header[1] = (long)Letter_Size;
      Header[2] = NLONGS;
      Header[3] = 0L;

      Payload[0] = Timing[0];
      Payload[1] = Timing[1];

      ctohl(Letter, NLONGS);
      G2_send(Letter, -1, -1);
   }

   Zip_pop();

#undef NLONGS
}

int timing_max_fn(long *ptr1, long *ptr2, int size, int nitems)
{
   if (size == 2 * sizeof(long))
   {
      if (ptr2[0] > ptr1[0])
      {
         ptr1[0] = ptr2[0];
         ptr1[1] = ptr2[1];
      }
      else if (ptr2[0] == ptr1[0])
      {
         if (ptr2[1] > ptr1[1])
            ptr1[1] = ptr2[1];
      }
   }
   else
   {
      print("timing_max_fn(): illegal size = %d vs %d", size, 2 * sizeof(long));
      exit(0);
   }
}

void receive_timing_letter(ZIP_MAILER *Mailer, long *sec, long *usec)
{
   char   *Subroutine_name = "receive_timing_letter():";
   int     p;
   int     q;

   char   *Letter;

   long   *Header;
   long   *Payload;

   Zip_push(Mailer);

   G2_pq(p, q);

   Letter = (char *) G2_recvb(0, 0);

   Header = (long *)Letter;

   if (*Header != TIMING_LETTER)
   {
      print("csparsetest1a: anticipated TIMING_LETTER=%ld, received %ld",
            (long)TIMING_LETTER, *Header);

      exit(1);
   }

   Payload = Header + 4;

   *sec  = Payload[0];
   *usec = Payload[1];

   Zip_free(Letter);

   Zip_pop();
}
